﻿Imports System.Runtime.CompilerServices
Imports System.Globalization
Imports System.ComponentModel

Namespace System

   Public Module IComparableX

      <Extension()> _
      Public Sub Maximize(Of T As IComparable)(ByRef subj As T, other As T)
         If other.CompareTo(subj) > 0 Then subj = other
      End Sub

      <Extension()> _
      Public Sub Minimize(Of T As IComparable)(ByRef subj As T, other As T)
         If other.CompareTo(subj) < 0 Then subj = other
      End Sub

      <Extension()> _
      Public Function IsInRange(Of T As IComparable)(toTest As T, lBound As T, uBound As T) As Boolean
         Return lBound.CompareTo(toTest) <= 0 AndAlso toTest.CompareTo(uBound) <= 0
      End Function
      <Extension()> _
      Public Function IsInRange(Of T As IComparable)(toTest As T, range As IEnumerable(Of T)) As Boolean
         Using enrt = range.GetEnumerator
            enrt.MoveNext()
            If enrt.Current.CompareTo(toTest) > 0 Then
               If Not enrt.MoveNext() Then Throw New ArgumentOutOfRangeException("range", range.Count(), "range must contain exactly 2 Elements")
               Return False
            End If
            enrt.MoveNext()
            IsInRange = enrt.Current.CompareTo(toTest) >= 0
            If enrt.MoveNext Then Throw New ArgumentOutOfRangeException("range", range.Count(), "range must contain exactly 2 Elements")
         End Using
      End Function

      <Extension()> _
      Public Function IsInRange(Of T As IComparable)( _
             toTest As T, _
             lBound As T, _
             uBound As T, _
             autoSort As Boolean, _
             lBoundInclude As Boolean, _
             uBoundInclude As Boolean) As Boolean
         If autoSort AndAlso lBound.CompareTo(uBound) > 0 Then
            Dim Temp = lBound
            lBound = uBound
            uBound = Temp
         End If
         Dim C0 As Integer = lBound.CompareTo(toTest)
         Dim C1 As Integer = toTest.CompareTo(uBound)
         If lBoundInclude AndAlso C0 = 0 Then Return True
         If UBoundInclude AndAlso C1 = 0 Then Return True
         Return C0 < 0 AndAlso C1 < 0
      End Function

      <Extension()> _
      Public Function Intersect(Of T As IComparable)(Range0 As IList(Of T), Range1 As IList(Of T)) As T()
         'der größere der LBounds und der kleinere der UBounds
         Return New T() {If(Range0(0).CompareTo(Range1(0)) > 0, Range0(0), Range1(0)), _
                               If(Range0(1).CompareTo(Range1(1)) < 0, Range0(1), Range1(1))}
      End Function

		''' <summary>gibt die Teile der beiden Ranges zurück, die sich nicht überlappen. Bei nur Berühren Vereinigung der Ranges</summary>
		<Extension()>
		Public Function IntersectXOR(Of T As IComparable)(Range0 As IList(Of T), Range1 As IList(Of T)) As T()()
			Dim result As T()()
			Dim isec = Range0.Intersect(Range1)
			If isec(0).CompareTo(isec(1)) >= 0 Then ' bei keine Überlappung ist result zunächstmal der Input
				result = {New T() {Range0(0), Range0(1)}, New T() {Range1(0), Range1(1)}}
			Else
				result = {New T() {Range0(0), Range1(0)}, New T() {Range0(1), Range1(1)}}
				result.ForEach(AddressOf Array.Sort)
				Select Case result(0)(0).CompareTo(result(1)(0))
					Case Is > 0 : result.Swap(0, 1)
					Case 0 : If result(0)(1).CompareTo(result(1)(1)) > 0 Then result.Swap(0, 1)
				End Select
			End If
			If result(0)(1).CompareTo(result(1)(0)) = 0 Then result = {New T() {result(0)(0), result(1)(1)}} ' bei Berührung vereinigen
			Return result.Where(Function(rng) rng(0).CompareTo(rng(1)) <> 0).ToArray ' Ranges der Breite 0 entfernen
		End Function

		Public Enum IntersectResult : No : Touch : Yes : YesOrTouch : End Enum
      <Extension()> _
      Public Function IsIntersectResult(Of T As IComparable)(Range() As T, query As IntersectResult) As Boolean
         Select Case query
            Case IntersectResult.No : Return Range(0).CompareTo(Range(1)) > 0
            Case IntersectResult.Touch : Return Range(0).CompareTo(Range(1)) = 0
            Case IntersectResult.Yes : Return Range(0).CompareTo(Range(1)) < 0
            Case IntersectResult.YesOrTouch : Return Range(0).CompareTo(Range(1)) <= 0
         End Select
      End Function

      <Extension()> _
      Public Function ClipIn(Of T As IComparable)(itm As T, LBound As T, UBound As T) As T
         If itm.CompareTo(LBound) < 0 Then Return LBound
         If itm.CompareTo(UBound) > 0 Then Return UBound
         Return itm
      End Function

      <Extension()> _
      Public Function Min(Of T As IComparable)(subj As T, other As T) As T
         If subj.CompareTo(other) < 0 Then Return subj
         Return other
      End Function
      <Extension()> _
      Public Function Max(Of T As IComparable)(subj As T, other As T) As T
         If subj.CompareTo(other) > 0 Then Return subj
         Return other
      End Function

      ''' <summary>
      ''' konvertiert eine Comparison in ein IComparer-implementierendes Objekt
      ''' </summary>
      ''' <remarks>selbst einen Delegaten kann man also erweitern</remarks>
      <Extension()> _
      Public Function ToComparer(Of T)(Subj As Comparison(Of T)) As ComparisonComparer(Of T)
         Return New ComparisonComparer(Of T)(Subj)
      End Function

      Public Class StringKeyComparer(Of T) : Inherits ComparisonComparer(Of T)
         Public Sub New(selector As Func(Of T, String), Optional options As CompareOptions = CompareOptions.StringSort Or CompareOptions.IgnoreCase)
            MyBase.New(Function(x As T, y As T) String.Compare(selector(x), selector(y), CultureInfo.InvariantCulture, options))
         End Sub
      End Class

      ''' <summary>IComparer-implementierender Wrapper um eine Comparison </summary>
      Public Class ComparisonComparer(Of T) : Inherits Comparer(Of T) : Implements IEqualityComparer(Of T)

         Private _Comparison As Comparison(Of T)
         Private _GetHashCode As Func(Of T, Integer)
         Public Sub New(Comparison As Comparison(Of T), Optional getHashCode As Func(Of T, Integer) = Nothing)
            _Comparison = Comparison
            _GetHashCode = getHashCode
         End Sub
         Public Overrides Function Compare(x As T, y As T) As Integer
            Return _Comparison(x, y)
         End Function

         Public Overloads Function Equals(x As T, y As T) As Boolean Implements IEqualityComparer(Of T).Equals
            Return _Comparison(x, y) = 0
         End Function

         Public Overloads Function GetHashCode(obj As T) As Integer Implements IEqualityComparer(Of T).GetHashCode
            Return _GetHashCode(obj)
         End Function
      End Class
   End Module
   Public Class Comparison2Comparer
      Public Shared Function FromSelector(Of T, TProp As IComparable(Of TProp))(selector As Func(Of T, TProp), Optional sortDirection As ListSortDirection = ListSortDirection.Ascending) As ComparisonComparer(Of T)
         Return New ComparisonComparer(Of T)( _
               If(sortDirection = ListSortDirection.Ascending, _
               Function(x As T, y As T) selector(x).CompareTo(selector(y)), _
               Function(x As T, y As T) selector(y).CompareTo(selector(x))), _
               Function(obj) selector(obj).GetHashCode)
      End Function
   End Class

End Namespace
